from __future__ import annotations

import glob
import json
import logging
import os
import time

import hdf5plugin
import numpy as np
from tqdm.auto import tqdm

from pygama import lgdo
from pygama.lgdo.lh5_store import DEFAULT_HDF5_COMPRESSION
from pygama.math.utils import sizeof_fmt

from .compass.compass_streamer import CompassStreamer
from .fc.fc_streamer import FCStreamer
from .orca.orca_streamer import OrcaStreamer
from .raw_buffer import RawBufferLibrary, write_to_lh5_and_clear

log = logging.getLogger(__name__)


def build_raw(
    in_stream: int,
    in_stream_type: str = None,
    out_spec: str | dict | RawBufferLibrary = None,
    buffer_size: int = 8192,
    n_max: int = np.inf,
    overwrite: bool = False,
    compass_config_file: str = None,
    hdf5_compression: str | dict | hdf5plugin.filters.Filter = DEFAULT_HDF5_COMPRESSION,
    **kwargs,
) -> None:
    """Convert data into LEGEND HDF5 raw-tier format.

    Takes an input stream of a given type and writes to output file(s)
    according to the user's a specification.

    Parameters
    ----------
    in_stream
        the name of the input stream to be converted. Typically a filename,
        including path. Can use environment variables. Some streamers may be
        able to (eventually) accept e.g. streaming over a port as an input.

    in_stream_type : 'ORCA', 'FlashCam', 'LlamaDaq', 'Compass' or 'MGDO'
        type of stream used to write the input file.

    out_spec
        Specification for the output stream.

        - if None, uses ``{in_stream}.lh5`` as the output filename.
        - if a str not ending in ``.json``, interpreted as the output filename.
        - if a str ending in ``.json``, interpreted as a filename containing
          json-shorthand for the output specification (see :mod:`.raw_buffer`).
        - if a JSON dict, should be a dict loaded from the json shorthand
          notation for RawBufferLibraries (see :mod:`.raw_buffer`), which is
          then used to build a :class:`.RawBufferLibrary`.
        - if a :class:`.RawBufferLibrary`, the mapping of data to output file /
          group is taken from that.

    buffer_size
        default size to use for data buffering.

    n_max
        maximum number of rows of data to process from the input file.

    overwrite
        sets whether to overwrite the output file(s) if it (they) already exist.

    compass_config_file
        specification of config file, used for decoding CoMPASS files

        - if None, CompassDecoder will sacrifice the first packet to determine
          waveform length
        - if a str ending in ``.json``, interpreted as a filename containing
          json-shorthand for the output specification (see
          :mod:`.compass.compass_event_decoder`).

    hdf5_compression
        forwarded to :meth:`~.lgdo.lh5_store.LH5Store.write_object`.

    **kwargs
        sent to :class:`.RawBufferLibrary` generation as `kw_dict` argument.
    """
    # convert any environment variables in in_stream so that we can check for readability
    in_stream = os.path.expandvars(in_stream)
    # later: fix if in_stream is not a file
    # (e.g. a socket, which exists if open and has size = np.inf)
    if not os.path.exists(in_stream):
        raise FileNotFoundError(f"file {in_stream} not found")

    in_stream_size = os.stat(in_stream).st_size

    # try to guess the input stream type if it's not provided
    if in_stream_type is None:
        i_ext = in_stream.split("/")[-1].rfind(".")
        if compass_config_file is not None:  # skip right away if compass
            in_stream_type = "Compass"
        elif i_ext != -1:
            ext = in_stream.split("/")[-1][i_ext + 1 :]
            if ext == "fcio":
                in_stream_type = "FlashCam"
            elif ext == "orca":
                in_stream_type = "ORCA"
            elif ext == "bin" or ext == "BIN":
                in_stream_type = "Compass"
            else:
                raise RuntimeError(
                    f"unknown file extension {ext}. Specify in_stream_type"
                )
        else:
            if OrcaStreamer.is_orca_stream(in_stream):  # orca default is no ext
                in_stream_type = "ORCA"
            else:
                raise RuntimeError("unknown file type. Specify in_stream_type")

    # process out_spec and setup rb_lib if specified
    rb_lib = None
    if isinstance(out_spec, str) and out_spec.endswith(".json"):
        with open(out_spec) as json_file:
            out_spec = json.load(json_file)
    if isinstance(out_spec, dict):
        out_spec = RawBufferLibrary(json_dict=out_spec, kw_dict=kwargs)
    if isinstance(out_spec, RawBufferLibrary):
        rb_lib = out_spec
    # if no rb_lib, write all data to file
    if out_spec is None:
        out_spec = in_stream
        i_ext = out_spec.rfind(".")
        if i_ext != -1:
            out_spec = out_spec[:i_ext]
        out_spec += ".lh5"
    # by now, out_spec should be a str or a RawBufferLibrary
    if not isinstance(out_spec, str) and not isinstance(out_spec, RawBufferLibrary):
        raise TypeError(f"unknown out_spec type {type(out_spec).__name__}")

    # modify buffer_size if necessary for n_max
    if buffer_size < 1:
        raise ValueError(f"bad buffer_size {buffer_size}")
    if buffer_size > n_max:
        buffer_size = n_max

    log.info(f"input: {in_stream}")
    out_files = [out_spec]
    if isinstance(out_spec, RawBufferLibrary):
        out_files = out_spec.get_list_of("out_stream")
    if len(out_files) == 1:
        log.info(f"output: {out_files[0]}")
    else:
        log.info("output:")
        for out_file in out_files:
            log.info(f" -> {out_file}")
    log.info(f"buffer size: {buffer_size}")
    if n_max < np.inf:
        log.info(f"maximum number of events: {n_max}")
    if log.getEffectiveLevel() <= logging.INFO:
        if n_max < np.inf:
            progress_bar = tqdm(desc="Decoding", total=n_max, delay=2, unit=" rows")
        else:
            progress_bar = tqdm(
                desc="Decoding",
                total=in_stream_size,
                delay=2,
                unit=" B",
                unit_scale=True,
            )

    # start a timer and a byte counter
    t_start = time.time()

    # select the appropriate streamer for in_stream
    streamer = None
    if in_stream_type == "ORCA":
        streamer = OrcaStreamer()
    elif in_stream_type == "FlashCam":
        streamer = FCStreamer()
    elif in_stream_type == "Compass":
        streamer = CompassStreamer(compass_config_file)
    elif in_stream_type == "LlamaDaq":
        raise NotImplementedError("LlamaDaq streaming not yet implemented")
    elif in_stream_type == "MGDO":
        raise NotImplementedError("MGDO streaming not yet implemented")
    else:
        raise NotImplementedError("unknown input stream type {in_stream_type}")

    # initialize the stream and read header. Also initializes rb_lib
    if log.getEffectiveLevel() <= logging.INFO:
        progress_bar.update(0)

    out_stream = out_spec if isinstance(out_spec, str) else ""
    header_data = streamer.open_stream(
        in_stream,
        rb_lib=rb_lib,
        buffer_size=buffer_size,
        chunk_mode="full_only",
        out_stream=out_stream,
    )
    rb_lib = streamer.rb_lib
    if log.getEffectiveLevel() <= logging.INFO and n_max == np.inf:
        progress_bar.update(streamer.n_bytes_read)

    # rb_lib should now be fully initialized. Check if files need to be
    # overwritten or if we need to stop to avoid overwriting
    out_files = rb_lib.get_list_of("out_stream")
    for out_file in out_files:
        colpos = out_file.find(":")
        if colpos != -1:
            out_file = out_file[:colpos]
        out_file_glob = glob.glob(out_file)
        if len(out_file_glob) == 0:
            continue
        if len(out_file_glob) > 1:
            raise RuntimeError(
                f"got multiple matches for out_file {out_file}: {out_file_glob}"
            )
        if not overwrite:
            raise FileExistsError(
                f"file {out_file_glob[0]} exists. Use option overwrite to proceed."
            )

        os.remove(out_file_glob[0])

    # Write header data
    lh5_store = lgdo.LH5Store(keep_open=True)
    write_to_lh5_and_clear(header_data, lh5_store, hdf5_compression=hdf5_compression)

    # Now loop through the data
    n_bytes_last = streamer.n_bytes_read
    while True:
        chunk_list = streamer.read_chunk()

        if log.getEffectiveLevel() <= logging.INFO and n_max == np.inf:
            progress_bar.update(streamer.n_bytes_read - n_bytes_last)
            n_bytes_last = streamer.n_bytes_read

        if len(chunk_list) == 0:
            log.debug("Reached EOF, exiting...")
            break

        n_read = 0
        for rb in chunk_list:
            if rb.loc > n_max:
                rb.loc = n_max
            n_max -= rb.loc
            n_read += rb.loc

        if log.getEffectiveLevel() <= logging.INFO and n_max < np.inf:
            progress_bar.update(n_read)

        write_to_lh5_and_clear(chunk_list, lh5_store, hdf5_compression=hdf5_compression)

        if n_max <= 0:
            log.info(f"Wrote {n_max} rows, exiting...")
            break

    streamer.close_stream()
    if log.getEffectiveLevel() <= logging.INFO:
        progress_bar.close()

    out_files = rb_lib.get_list_of("out_stream")
    if len(out_files) == 1:
        out_file = out_files[0].split(":", 1)[0]
        if os.path.exists(out_file):
            file_size = os.stat(out_file).st_size
            log.info(f"output file: {out_file} ({sizeof_fmt(file_size)})")
        else:
            log.info(f"output file: {out_file} (not written)")
    else:
        log.info("output files:")
        unique_outfilenames = {file.split(":", 1)[0] for file in out_files}
        for out_file in unique_outfilenames:
            if os.path.exists(out_file):
                file_size = os.stat(out_file).st_size
                log.info(f" -> {out_file} ({sizeof_fmt(file_size)})")
            else:
                log.info(f" -> {out_file} (not written)")
    log.info(f"total converted: {sizeof_fmt(streamer.n_bytes_read)}")
    elapsed = time.time() - t_start
    log.info(f"conversion speed: {sizeof_fmt(streamer.n_bytes_read/elapsed)}ps")



from __future__ import annotations

from typing import Callable

import numpy as np
from numba import guvectorize
from pyfftw import FFTW

from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


def dft(buf_in: np.ndarray, buf_out: np.ndarray) -> Callable:
    """Perform discrete Fourier transforms using the FFTW library.

    Note
    ----
    This processor is composed of a factory function that is called using the
    `init_args` argument. The input and output waveforms are passed using
    `args`.

    Note
    ----
    FFTW optimizes the FFT algorithm based on the size of the arrays, with SIMD
    parallelized commands.  This optimization requires initialization, so this
    is a factory function that returns a Numba gufunc that performs the FFT.
    FFTW works on fixed memory buffers, so you must tell it what memory to use
    ahead of time.  When using this with
    :class:`~.dsp.processing_chain.ProcessingChain`, to ensure the correct
    buffers are used, call
    :meth:`~.dsp.processing_chain.ProcessingChain.get_variable` to give it the
    internal memory buffer directly. With :func:`~.dsp.build_dsp.build_dsp`,
    you can just give it the name, and it will automatically happen. The
    possible `dtypes` for the input/outputs are:

    =============================== ========= =============================== =============
    :class:`numpy.dtype`            Size      :class:`numpy.dtype`            Size
    ------------------------------- --------- ------------------------------- -------------
    ``float32``/``float``           :math:`n` ``complex64``                   :math:`n/2+1`
    ``float64``/``double``          :math:`n` ``complex128``                  :math:`n/2+1`
    ``float128``/``longdouble``     :math:`n` ``complex256``/``clongdouble``  :math:`n/2+1`
    ``complex64``                   :math:`n` ``complex64``                   :math:`n`
    ``complex128``                  :math:`n` ``complex128``                  :math:`n`
    ``complex256``/``clongdouble``  :math:`n` ``complex256``/``clongdouble``  :math:`n`
    =============================== ========= =============================== =============
    """
    try:
        dft_fun = FFTW(buf_in, buf_out, axes=(-1,), direction="FFTW_FORWARD")
    except ValueError:
        raise ValueError(
            "incompatible array types/shapes. See function documentation for allowed values"
        )

    typesig = "void(" + str(buf_in.dtype) + "[:, :], " + str(buf_out.dtype) + "[:, :])"
    sizesig = "(m, n)->(m, n)" if buf_in.shape == buf_out.shape else "(m, n),(m, l)"

    @guvectorize(
        [typesig],
        sizesig,
        **nb_kwargs(
            cache=False,
            forceobj=True,
        ),
    )
    def dft(wf_in: np.ndarray, dft_out: np.ndarray) -> None:
        dft_fun(wf_in, dft_out)

    return dft


def inv_dft(buf_in: np.ndarray, buf_out: np.ndarray) -> Callable:
    """Perform inverse discrete Fourier transforms using the FFTW library.

    Note
    ----
    This processor is composed of a factory function that is called using the
    `init_args` argument. The input and output waveforms are passed using
    `args`.

    Note
    ----
    FFTW optimizes the FFT algorithm based on the size of the arrays, with SIMD
    parallelized commands.  This optimization requires initialization, so this
    is a factory function that returns a Numba gufunc that performs the FFT.
    FFTW works on fixed memory buffers, so you must tell it what memory to use
    ahead of time.  When using this with
    :class:`~.dsp.processing_chain.ProcessingChain`, to ensure the correct
    buffers are used, call
    :meth:`~.dsp.processing_chain.ProcessingChain.get_variable` to give it the
    internal memory buffer directly. With :func:`~.dsp.build_dsp.build_dsp`,
    you can just give it the name, and it will automatically happen. The
    possible `dtypes` for the input/outputs are:

    =============================== ============= =============================== =========
    :class:`numpy.dtype`            Size          :class:`numpy.dtype`            Size
    ------------------------------- ------------- ------------------------------- ---------
    ``complex64``                   :math:`n/2+1` ``float32``/``float``           :math:`n`
    ``complex128``                  :math:`n/2+1` ``float64``/``double``          :math:`n`
    ``complex256``/``clongdouble``  :math:`n/2+1` ``float128``/``longdouble``     :math:`n`
    ``complex64``                   :math:`n`     ``complex64``                   :math:`n`
    ``complex128``                  :math:`n`     ``complex128``                  :math:`n`
    ``complex256``/``clongdouble``  :math:`n`     ``complex256``/``clongdouble``  :math:`n`
    =============================== ============= =============================== =========
    """
    try:
        idft_fun = FFTW(buf_in, buf_out, axes=(-1,), direction="FFTW_BACKWARD")
    except ValueError:
        raise ValueError(
            "incompatible array types/shapes. See function documentation for allowed values"
        )

    typesig = "void(" + str(buf_in.dtype) + "[:, :], " + str(buf_out.dtype) + "[:, :])"
    sizesig = "(m, n)->(m, n)" if buf_in.shape == buf_out.shape else "(m, n),(m, l)"

    @guvectorize(
        [typesig],
        sizesig,
        **nb_kwargs(
            cache=False,
            forceobj=True,
        ),
    )
    def inv_dft(wf_in: np.ndarray, dft_out: np.ndarray) -> None:
        idft_fun(wf_in, dft_out)

    return inv_dft


def psd(buf_in: np.ndarray, buf_out: np.ndarray) -> Callable:
    """Perform discrete Fourier transforms using the FFTW library, and use it to get
    the power spectral density.

    Note
    ----
    This processor is composed of a factory function that is called using the
    `init_args` argument. The input and output waveforms are passed using
    `args`.

    Note
    ----
    FFTW optimizes the FFT algorithm based on the size of the arrays, with SIMD
    parallelized commands.  This optimization requires initialization, so this
    is a factory function that returns a Numba gufunc that performs the FFT.
    FFTW works on fixed memory buffers, so you must tell it what memory to use
    ahead of time.  When using this with
    :class:`~.dsp.processing_chain.ProcessingChain`, to ensure the correct
    buffers are used, call
    :meth:`~.dsp.processing_chain.ProcessingChain.get_variable` to give it the
    internal memory buffer directly. With :func:`~.dsp.build_dsp.build_dsp`,
    you can just give it the name, and it will automatically happen. The
    possible `dtypes` for the input/outputs are:

    =============================== ========= ============================ =============
    :class:`numpy.dtype`            Size      :class:`numpy.dtype`         Size
    ------------------------------- --------- ---------------------------- -------------
    ``complex64``                   :math:`n` ``float32``/``float``        :math:`n`
    ``complex128``                  :math:`n` ``float64``/``double``       :math:`n`
    ``complex256``/``clongdouble``  :math:`n` ``float128``/``longdouble``  :math:`n`
    ``float32``/``float``           :math:`n` ``float32``/``float``        :math:`n/2+1`
    ``float64``/``double``          :math:`n` ``float64``/``double``       :math:`n/2+1`
    ``float128``/``longdouble``     :math:`n` ``float128``/``longdouble``  :math:`n/2+1`
    =============================== ========= ============================ =============
    """

    # build intermediate array for the dft, which will be abs'd to get the PSD
    buf_dft = np.ndarray(
        buf_out.shape, np.dtype("complex" + str(buf_out.dtype.itemsize * 16))
    )
    try:
        dft_fun = FFTW(buf_in, buf_dft, axes=(-1,), direction="FFTW_FORWARD")
    except ValueError:
        raise ValueError(
            "incompatible array types/shapes. See function documentation for allowed values"
        )

    typesig = "void(" + str(buf_in.dtype) + "[:, :], " + str(buf_out.dtype) + "[:, :])"
    sizesig = "(m, n)->(m, n)" if buf_in.shape == buf_out.shape else "(m, n),(m, l)"

    @guvectorize(
        [typesig],
        sizesig,
        **nb_kwargs(
            cache=False,
            forceobj=True,
        ),
    )
    def psd(wf_in: np.ndarray, psd_out: np.ndarray) -> None:
        dft_fun(wf_in, buf_dft)
        np.abs(buf_dft, psd_out)

    return psd



import copy
import logging
import sys
from typing import Any

import numpy as np

from pygama.raw.orca.orca_base import OrcaDecoder, get_ccc
from pygama.raw.orca.orca_header import OrcaHeader
from pygama.raw.orca.orca_packet import OrcaPacket
from pygama.raw.raw_buffer import RawBufferLibrary

log = logging.getLogger(__name__)


class ORSIS3302DecoderForEnergy(OrcaDecoder):
    """Decoder for `Struck SIS3302 <https://www.struck.de/sis3302.htm>`_ digitizer
    data written by ORCA.
    """

    def __init__(self, header: OrcaHeader = None, **kwargs) -> None:
        self.decoded_values_template = {
            "packet_id": {
                "dtype": "uint32",
            },
            "energy": {
                "dtype": "uint32",
                "units": "adc",
            },
            "energy_first": {
                "dtype": "uint32",
            },
            "timestamp": {
                "dtype": "uint64",
                "units": "clock_ticks",
            },
            "crate": {
                "dtype": "uint8",
            },
            "card": {
                "dtype": "uint8",
            },
            "channel": {
                "dtype": "uint8",
            },
            "waveform": {
                "dtype": "uint16",
                "datatype": "waveform",
                "wf_len": 65532,  # max value. override this before initializing buffers to save RAM
                "dt": 10,  # override if a different clock rate is used
                "dt_units": "ns",
                "t0_units": "ns",
            },
        }
        """Default Struck SIS3302 (read out by ORCA) waveform decoded values.

        Warning
        -------
        This configuration can be dynamically modified by the decoder at runtime.
        """
        self.decoded_values = {}
        super().__init__(header=header, **kwargs)
        self.skipped_channels = {}

    def set_header(self, header: OrcaHeader) -> None:
        self.header = header

        obj_info_dict = header.get_object_info("ORSIS3302Model")
        # Loop over crates, cards, build decoded values for enabled channels
        for crate in obj_info_dict:
            for card in obj_info_dict[crate]:
                int_enabled_mask = obj_info_dict[crate][card][
                    "internalTriggerEnabledMask"
                ]
                ext_enabled_mask = obj_info_dict[crate][card][
                    "externalTriggerEnabledMask"
                ]
                enabled_mask = int_enabled_mask | ext_enabled_mask
                for channel in range(8):
                    # only care about enabled channels
                    if not ((enabled_mask >> channel) & 0x1):
                        continue

                    ccc = get_ccc(crate, card, channel)

                    self.decoded_values[ccc] = copy.deepcopy(
                        self.decoded_values_template
                    )

                    # get trace length(s). Should all be the same until
                    # multi-buffer mode is implemented AND each channel has its
                    # own buffer
                    trace_length = obj_info_dict[crate][card]["sampleLengths"][
                        int(channel / 2)
                    ]
                    if trace_length <= 0 or trace_length > 2**16:
                        raise RuntimeError(f"invalid trace_length {trace_length}")
                        sys.exit()
                    self.decoded_values[ccc]["waveform"]["wf_len"] = trace_length

    def get_key_lists(self) -> list[list[str]]:
        key_lists = []
        for key in self.decoded_values.keys():
            key_lists.append([key])
        return [key_lists]

    def get_decoded_values(self, key: int = None) -> dict[str, Any]:
        if key is None:
            dec_vals_list = self.decoded_values.values()
            if len(dec_vals_list) >= 0:
                return list(dec_vals_list)[0]
            raise RuntimeError("decoded_values not built")
        if key in self.decoded_values:
            return self.decoded_values[key]
        raise KeyError(f"no decoded values for key {key}")

    def decode_packet(
        self, packet: OrcaPacket, packet_id: int, rbl: RawBufferLibrary
    ) -> bool:
        evt_rbkd = rbl.get_keyed_dict()

        # read the crate/card/channel first
        crate = (packet[1] >> 21) & 0xF
        card = (packet[1] >> 16) & 0x1F
        channel = (packet[1] >> 8) & 0xFF
        ccc = get_ccc(crate, card, channel)

        # get the table for this crate/card/channel
        if ccc not in evt_rbkd:
            if ccc not in self.skipped_channels:
                self.skipped_channels[ccc] = 0
                log.debug(f"Skipping channel: {ccc}")
                log.debug(f"evt_rbkd: {evt_rbkd.keys()}")
            self.skipped_channels[ccc] += 1
            return False
        tbl = evt_rbkd[ccc].lgdo
        ii = evt_rbkd[ccc].loc

        # store packet id
        tbl["packet_id"].nda[ii] = packet_id

        # read the rest of the record
        tbl["crate"].nda[ii] = crate
        tbl["card"].nda[ii] = card
        tbl["channel"].nda[ii] = channel
        buffer_wrap = packet[1] & 0x1
        wf_length32 = packet[2]
        ene_wf_length32 = packet[3]
        tbl["timestamp"].nda[ii] = packet[5] + (
            (packet[4] & 0xFFFF0000) << 16
        )  # might need to convert to uint64
        last_word = packet[-1]

        # get the footer
        tbl["energy"].nda[ii] = packet[-4]
        tbl["energy_first"].nda[ii] = packet[-3]

        # Interpret the raw event data into numpy array of 16 bit ints.
        # Does not copy data. p16 is read-only
        p16 = np.frombuffer(packet, dtype=np.uint16)

        # compute expected and actual array dimensions
        wf_length16 = 2 * wf_length32
        ene_wf_length16 = 2 * ene_wf_length32
        orca_header_length16 = 8
        sis_header_length16 = 8 if buffer_wrap else 4
        header_length16 = orca_header_length16 + sis_header_length16
        footer_length16 = 8
        expected_wf_length16 = (
            len(p16) - header_length16 - footer_length16 - ene_wf_length16
        )

        # error check: waveform size must match expectations
        if wf_length16 != expected_wf_length16 or last_word != 0xDEADBEEF:
            raise RuntimeError(
                f"Waveform size {wf_length16} doesn't match expected size {expected_wf_length16}. "
                "The Last Word (should be 0xdeadbeef): {hex(last_word)}"
            )

        # splitting waveform indices into two chunks (all referring to the 16 bit array)
        i_wf_start = header_length16
        i_wf_stop = i_wf_start + wf_length16
        if buffer_wrap:
            # start somewhere in the middle of the record
            i_start_1 = packet[7] + header_length16 + 1
            i_stop_1 = i_wf_stop  # end of the wf record
            i_start_2 = i_wf_start  # beginning of the wf record
            i_stop_2 = i_start_1

        # handle the waveform(s)
        tbwf = tbl["waveform"]["values"].nda[ii]
        if wf_length32 > 0:
            if not buffer_wrap:
                if i_wf_stop - i_wf_start != expected_wf_length16:
                    raise RuntimeError(
                        f"We expected {expected_wf_length16} waveform samples "
                        "and only got {i_wf_stop-i_wf_start}"
                    )
                tbwf[:expected_wf_length16] = p16[i_wf_start:i_wf_stop]
            else:
                len1 = i_stop_1 - i_start_1
                len2 = i_stop_2 - i_start_2
                if len1 + len2 != expected_wf_length16:
                    raise RuntimeError(
                        f"We expected {expected_wf_length16} waveform samples and only got {len1+len2}"
                    )
                tbwf[:len1] = p16[i_start_1:i_stop_1]
                tbwf[len1 : len1 + len2] = p16[i_start_2:i_stop_2]

        evt_rbkd[ccc].loc += 1
        return evt_rbkd[ccc].is_full()


class ORSIS3316WaveformDecoder(OrcaDecoder):
    """Decoder for `Struck SIS3316 <https://www.struck.de/sis3316.html>`_
    digitizer data written by ORCA."""

    def __init__(self, header: OrcaHeader = None, **kwargs) -> None:
        # store an entry for every event
        self.decoded_values_template = {
            "packet_id": {"dtype": "uint32"},
            "energy": {"dtype": "uint32", "units": "adc"},
            "peakHighValue": {"dtype": "uint32", "units": "adc"},
            "peakHighIndex": {"dtype": "uint32", "units": "adc"},
            "accSum1": {"dtype": "uint32", "units": "adc"},
            "information": {"dtype": "uint32"},
            "accSum2": {"dtype": "uint32", "units": "adc"},
            "accSum3": {"dtype": "uint32", "units": "adc"},
            "accSum4": {"dtype": "uint32", "units": "adc"},
            "accSum5": {"dtype": "uint32", "units": "adc"},
            "accSum6": {"dtype": "uint32", "units": "adc"},
            "accSum7": {"dtype": "uint32", "units": "adc"},
            "accSum8": {"dtype": "uint32", "units": "adc"},
            "mawMax": {"dtype": "uint32", "units": "adc"},
            "mawBefore": {"dtype": "uint32", "units": "adc"},
            "mawAfter": {"dtype": "uint32", "units": "adc"},
            "startEnergy": {"dtype": "uint32", "units": "adc"},
            "maxEnergy": {"dtype": "uint32", "units": "adc"},
            "energy_first": {"dtype": "uint32"},
            "timestamp": {"dtype": "uint64", "units": "clock_ticks"},
            "crate": {"dtype": "uint8"},
            "card": {"dtype": "uint8"},
            "channel": {"dtype": "uint8"},
            # waveform data
            "waveform": {
                "dtype": "uint16",
                "datatype": "waveform",
                "wf_len": 65532,  # max value. override this before initializing buffers to save RAM
                "dt": 8,  # override if a different clock rate is used
                "dt_units": "ns",
                "t0_units": "ns",
            },
        }
        """Default Struck SIS3316 (read out by ORCA) waveform decoded values.

        Warning
        -------
        This configuration can be dynamically modified by the decoder at runtime.
        """

        self.decoded_values = {}
        self.skipped_channels = {}
        super().__init__(
            header=header, **kwargs
        )  # also initializes the garbage df (whatever that means...)

    def set_header(self, header: OrcaHeader) -> None:
        self.header = header

        for card_dict in self.header["ObjectInfo"]["Crates"][0]["Cards"]:
            if card_dict["Class Name"] == "ORSIS3316Model":
                channel_list = [int(d) for d in str(bin(card_dict["enabledMask"]))[2:]]
                card = card_dict["Card"]
                crate = 0
                for i in range(0, len(channel_list)):
                    if int(channel_list[i]) == 1:
                        channel = i
                        ccc = get_ccc(crate, card, channel)
                        trace_length = card_dict["rawDataBufferLen"]
                        self.decoded_values[ccc] = copy.deepcopy(
                            self.decoded_values_template
                        )

                        if trace_length <= 0 or trace_length > 2**16:
                            raise RuntimeError(
                                "invalid trace_length: ",
                                trace_length,
                            )

                        self.decoded_values[ccc]["waveform"]["wf_len"] = trace_length
                    else:
                        continue

    def get_key_lists(self) -> list[list[str]]:
        key_lists = []
        for key in self.decoded_values.keys():
            key_lists.append([key])
        return [key_lists]

    def get_decoded_values(self, key: int = None) -> dict[str, Any]:
        if key is None:
            dec_vals_list = self.decoded_values.values()
            if len(dec_vals_list) == 0:
                raise RuntimeError("decoded_values not built yet!")

            return dec_vals_list  # Get first thing we find
        else:
            dec_vals_list = self.decoded_values[key]
            return dec_vals_list
        raise RuntimeError("no decoded values for key", key)

    def decode_packet(
        self, packet: OrcaPacket, packet_id: int, rbl: RawBufferLibrary
    ) -> bool:
        """Decode the ORCA SIS3316 ADC packet.

        The packet is formatted as:

        .. code-block:: text

            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx
            ^^^^ ^^^^ ^^^^ ^^----------------------- Data ID (from header)
            -----------------^^ ^^^^ ^^^^ ^^^^ ^^^^- length
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx
            --------^-^^^--------------------------- Crate number
            -------------^-^^^^--------------------- Card number
            --------------------^^^^ ^^^^----------- Chan number
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx  Num Events in this packet
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx  Num longs in each record
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx  Num of Records that were in the FIFO
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx  Num of longs in data header -- can get from the raw data also
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx  Spare
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx  Spare
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx  Spare
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx  Spare

        Then, data events follow with format described in below (note the format bits):

        .. code-block:: text

            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx
            -----------------------------------^^^^- Format bits (from header)
            --------------------^^^^ ^^^^ ^^^^------ Channel ID
            ^^^^ ^^^^ ^^^^ ^^^^--------------------- Timestamp[47:32]
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx  Timestamp[31:0]
            if Format bit 0 = 1 add
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx
            --------------------^^^^ ^^^^ ^^^^ ^^^^- Peakhigh value
            ^^^^ ^^^^ ^^^^ ^^^^--------------------- Index of Peakhigh value
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx
            ----------^^^^ ^^^^ ^^^^ ^^^^ ^^^^ ^^^^- Accumulator sum of Gate 1 [23:0]
            ^^^^ ^^^^------------------------------- Information byte
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx Accumulator sum of Gate 2 [27:0]
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx Accumulator sum of Gate 3 [27:0]
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx Accumulator sum of Gate 4 [27:0]
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx Accumulator sum of Gate 5 [27:0]
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx Accumulator sum of Gate 6 [27:0]
            If Format bit 1 = 1 add
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx Accumulator sum of Gate 7 [27:0]
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx Accumulator sum of Gate 8 [27:0]
            If Format bit 2 = 1 add
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx MAW Maximum value [27:0]
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx MAW Value before Trigger [27:0]
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx MAW Value after/with Trigger [27:0]
            If Format bit 3 = 1 add
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx Start Energy Value (in Trigger Gate)
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx Max Energy Value (in Trigger Gate)
            Regardless of format bit
            xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx
            -------^^ ^^^^ ^^^^ ^^^^ ^^^^ ^^^^ ^^^^- number of raw samples divided by 2
            ------^--------------------------------- status Flag
            -----^---------------------------------- MAW Test Flag
            Followed by N ADC Raw Samples (2 Samples per 32 bit word)
            Followed by MAW Test data
        """

        evt_rbkd = rbl.get_keyed_dict()

        evt_data_16 = np.frombuffer(packet, dtype=np.uint16)

        # First grab the crate, card, and channel to calculate ccc
        crate = (packet[1] >> 21) & 0xF
        card = (packet[1] >> 16) & 0x1F
        channel = (packet[1] >> 8) & 0xFF
        ccc = get_ccc(crate, card, channel)

        # Check if this ccc should be recorded.
        if ccc not in evt_rbkd:
            if ccc not in self.skipped_channels:
                self.skipped_channels[ccc] = 0
                log.debug(f"Skipping channel: {ccc}")
                log.debug(f"evt_rbkd: {evt_rbkd.keys()}")
            self.skipped_channels[ccc] += 1
            return False

        orca_header_length = 10
        # Find the number of Events that is in this packet.
        num_of_events = packet[2]
        num_of_longs = packet[3]
        data_header_length = packet[5]

        # Creating the first table and finding the index.
        tbl = evt_rbkd[ccc].lgdo

        for i in range(0, num_of_events):
            # will crash if at the end of buffer.
            ii = evt_rbkd[ccc].loc
            if ii <= 8191:
                # save the crate, card, and channel number which does not change
                tbl["crate"].nda[ii] = crate
                tbl["card"].nda[ii] = card
                tbl["channel"].nda[ii] = channel

                # calculates where to start indexing based on the num_of_longs
                # for each event offset by the orca_header_length.
                event_start = orca_header_length + num_of_longs * i

                if len(packet) > 10:
                    tbl["timestamp"].nda[ii] = packet[event_start + 1] + (
                        (packet[event_start] & 0xFFFF0000) << 16
                    )
                else:
                    tbl["timestamp"].nda[ii] = 0

                index = event_start + 2
                if (packet[event_start] & 0x1) == 1:
                    tbl["peakHighValue"].nda[ii] = packet[index] & 0xFFFF
                    tbl["peakHighIndex"].nda[ii] = (packet[index] >> 16) & 0xFFFF
                    index += 1
                    tbl["accSum1"].nda[ii] = packet[index] & 0xFFFFFF
                    tbl["information"].nda[ii] = (packet[index] >> 24) & 0xFF
                    index += 1
                    tbl["accSum2"].nda[ii] = packet[index]
                    index += 1
                    tbl["accSum3"].nda[ii] = packet[index]
                    index += 1
                    tbl["accSum4"].nda[ii] = packet[index]
                    index += 1
                    tbl["accSum5"].nda[ii] = packet[index]
                    index += 1
                    tbl["accSum6"].nda[ii] = packet[index]
                else:
                    tbl["peakHighValue"].nda[ii] = -1
                    tbl["peakHighIndex"].nda[ii] = -1
                    tbl["accSum1"].nda[ii] = -1
                    tbl["accSum2"].nda[ii] = -1
                    tbl["accSum3"].nda[ii]
                    tbl["accSum4"].nda[ii]
                    tbl["accSum5"].nda[ii] = -1
                    tbl["accSum6"].nda[ii] = -1

                if ((packet[event_start] >> 1) & 0x1) == 1:
                    tbl["accSum7"].nda[ii] = packet[index]
                    index += 1
                    tbl["accSum8"].nda[ii] = packet[index]
                    index += 1
                else:
                    tbl["accSum7"].nda[ii] = -1
                    tbl["accSum8"].nda[ii] = -1

                if ((packet[event_start] >> 2) & 0x1) == 1:
                    tbl["mawMax"].nda[ii] = packet[index]
                    index += 1
                    tbl["mawBefore"].nda[ii] = packet[index]
                    index += 1
                    tbl["mawAfter"].nda[ii] = packet[index]
                    index += 1
                else:
                    tbl["mawMax"].nda[ii] = -1
                    tbl["mawBefore"].nda[ii] = -1
                    tbl["mawAfter"].nda[ii] = -1

                if ((packet[event_start] >> 3) & 0x1) == 1:
                    tbl["startEnergy"].nda[ii] = packet[index]
                    index += 1
                    tbl["maxEnergy"].nda[ii] = packet[index]
                    index += 1
                else:
                    tbl["startEnergy"].nda[ii] = -1
                    tbl["maxEnergy"].nda[ii] = -1

                data_header_length16 = data_header_length * 2

                expected_wf_length = num_of_longs * 2 - data_header_length16

                i_wf_start = data_header_length16 + event_start * 2

                i_wf_stop = i_wf_start + expected_wf_length

                if expected_wf_length > 0:
                    tbl["waveform"]["values"].nda[ii] = evt_data_16[
                        i_wf_start:i_wf_stop
                    ]

                # move to the next index for the next event.
                evt_rbkd[ccc].loc += 1

        return evt_rbkd[ccc].is_full()



import numpy as np
import pytest

from pygama import lgdo
from pygama.lgdo import VectorOfVectors


@pytest.fixture()
def lgdo_vov():
    return VectorOfVectors(
        flattened_data=lgdo.Array(
            nda=np.array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1])
        ),
        cumulative_length=lgdo.Array(nda=np.array([2, 5, 6, 10, 13])),
    )
    # [1 2]
    # [3 4 5]
    # [2]
    # [4 8 9 7]
    # [5 3 1]


def test_init(lgdo_vov):
    assert len(VectorOfVectors(dtype="ubyte")) == 0

    vov = VectorOfVectors(shape_guess=(10, 20), dtype="int32", fill_val=2)
    assert vov.flattened_data == lgdo.Array(shape=(10 * 20,), fill_val=2, dtype="int32")
    assert vov.cumulative_length == lgdo.Array(
        np.arange(20, 10 * 20 + 1, 20, dtype="uint32")
    )

    test = VectorOfVectors([[1, 2], [3, 4, 5], [2], [4, 8, 9, 7], [5, 3, 1]])
    assert test == lgdo_vov
    assert len(test) == 5

    vov = VectorOfVectors(
        cumulative_length=np.array([5, 10, 15], dtype="uint32"), dtype="ubyte"
    )
    assert len(vov.flattened_data) == 15
    assert len(vov[-1]) == 5

    vov = VectorOfVectors(shape_guess=(5, 0), dtype="int32")
    assert vov.cumulative_length == lgdo.Array([0, 0, 0, 0, 0])


def test_datatype_name(lgdo_vov):
    assert lgdo_vov.datatype_name() == "array"


def test_form_datatype(lgdo_vov):
    assert lgdo_vov.form_datatype() == "array<1>{array<1>{real}}"


def test_values(lgdo_vov):
    desired = [
        np.array([1, 2]),
        np.array([3, 4, 5]),
        np.array([2]),
        np.array([4, 8, 9, 7]),
        np.array([5, 3, 1]),
    ]

    for i in range(len(desired)):
        assert np.array_equal(desired[i], lgdo_vov[i])

    assert np.array_equal(lgdo_vov[-1], desired[-1])
    assert np.array_equal(lgdo_vov[-2], desired[-2])

    vov = VectorOfVectors([[1, 2]], dtype="uint32")
    assert np.array_equal(vov[-1], [1, 2])


def test_resize(lgdo_vov):
    lgdo_vov.resize(3)
    assert len(lgdo_vov.cumulative_length) == 3
    assert len(lgdo_vov.flattened_data) == lgdo_vov.cumulative_length[-1]

    desired = [np.array([1, 2]), np.array([3, 4, 5]), np.array([2])]

    for i in range(len(lgdo_vov)):
        assert np.array_equal(desired[i], lgdo_vov[i])

    lgdo_vov.resize(5)
    assert len(lgdo_vov) == 5
    assert len(lgdo_vov[3]) == 0
    assert len(lgdo_vov[4]) == 0
    assert lgdo_vov == VectorOfVectors([[1, 2], [3, 4, 5], [2], [], []])

    vov = VectorOfVectors(dtype="i")
    vov.resize(3)
    assert vov == VectorOfVectors([[], [], []], dtype="i")


def test_aoesa(lgdo_vov):
    arr = lgdo_vov.to_aoesa()
    desired = np.array(
        [
            [1, 2, np.nan, np.nan],
            [3, 4, 5, np.nan],
            [2, np.nan, np.nan, np.nan],
            [4, 8, 9, 7],
            [5, 3, 1, np.nan],
        ]
    )
    assert isinstance(arr, lgdo.ArrayOfEqualSizedArrays)
    assert arr.dtype == np.float64
    assert np.array_equal(arr.nda, desired, True)

    vov = VectorOfVectors(
        flattened_data=lgdo.Array(
            nda=np.array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1], dtype="int16")
        ),
        cumulative_length=lgdo.Array(nda=np.array([2, 5, 6, 10, 13])),
    )
    aoesa = vov.to_aoesa()
    assert aoesa.dtype == np.float64

    aoesa = vov.to_aoesa(preserve_dtype=True)
    assert aoesa.dtype == np.int16


def test_set_vector(lgdo_vov):
    lgdo_vov[0] = np.zeros(2)

    desired = [
        np.zeros(2),
        np.array([3, 4, 5]),
        np.array([2]),
        np.array([4, 8, 9, 7]),
        np.array([5, 3, 1]),
    ]

    for i in range(len(desired)):
        assert np.array_equal(desired[i], lgdo_vov[i])

    with pytest.raises(ValueError):
        lgdo_vov[0] = np.zeros(3)

    lgdo_vov[1] = np.zeros(3)

    desired = [
        np.zeros(2),
        np.zeros(3),
        np.array([2]),
        np.array([4, 8, 9, 7]),
        np.array([5, 3, 1]),
    ]

    for i in range(len(desired)):
        assert np.array_equal(desired[i], lgdo_vov[i])


def test_append(lgdo_vov):
    lgdo_vov.append(np.zeros(3))
    assert np.array_equal(lgdo_vov[-1], np.zeros(3))

    vov = VectorOfVectors(dtype="int64")
    vov.append(np.zeros(3))
    assert vov == VectorOfVectors([[0, 0, 0]])


def test_insert(lgdo_vov):
    lgdo_vov.insert(2, np.zeros(3))
    assert lgdo_vov == VectorOfVectors(
        [
            [1, 2],
            [3, 4, 5],
            [0, 0, 0],
            [2],
            [4, 8, 9, 7],
            [5, 3, 1],
        ]
    )

    vov = VectorOfVectors(shape_guess=(3, 5), dtype="int32", fill_val=0)
    vov.insert(2, [1, 2, 3])
    assert np.array_equal(vov.cumulative_length, [5, 10, 13, 18])
    assert np.array_equal(vov[2], [1, 2, 3])


def test_replace(lgdo_vov):
    vov = lgdo.copy(lgdo_vov)
    vov.replace(1, np.zeros(3))
    assert vov == VectorOfVectors(
        [
            [1, 2],
            [0, 0, 0],
            [2],
            [4, 8, 9, 7],
            [5, 3, 1],
        ]
    )

    vov = lgdo.copy(lgdo_vov)
    vov.replace(1, np.zeros(2))
    assert vov == VectorOfVectors(
        [
            [1, 2],
            [0, 0],
            [2],
            [4, 8, 9, 7],
            [5, 3, 1],
        ]
    )

    vov = lgdo.copy(lgdo_vov)
    vov.replace(1, np.zeros(4))
    assert vov == VectorOfVectors(
        [
            [1, 2],
            [0, 0, 0, 0],
            [2],
            [4, 8, 9, 7],
            [5, 3, 1],
        ]
    )


def test_iter(lgdo_vov):
    desired = [
        np.array([1, 2]),
        np.array([3, 4, 5]),
        np.array([2]),
        np.array([4, 8, 9, 7]),
        np.array([5, 3, 1]),
    ]

    c = 0
    for v in lgdo_vov:
        assert (v == desired[c]).all()
        c += 1


def test_build_cl_and_explodes():
    cl = np.array([3, 4], dtype=np.uint64)
    exp = np.array([0, 0, 0, 1], dtype=np.uint64)
    array = np.array([5, 7], dtype=np.uint64)
    array_exp = np.array([5, 5, 5, 7], dtype=np.uint64)
    # build_cl
    assert (lgdo.build_cl(exp, cl) == cl).all()
    assert (lgdo.build_cl(exp) == cl).all()
    assert (lgdo.build_cl([0, 0, 0, 1]) == cl).all()
    assert (lgdo.build_cl(array_exp, cl) == cl).all()
    assert (lgdo.build_cl(array_exp) == cl).all()
    assert (lgdo.build_cl([5, 5, 5, 7]) == cl).all()
    # explode_cl
    assert (lgdo.explode_cl(cl, exp) == exp).all()
    assert (lgdo.explode_cl(cl) == exp).all()
    assert (lgdo.explode_cl([3, 4]) == exp).all()
    # inverse functionality
    assert (lgdo.build_cl(lgdo.explode_cl(cl)) == cl).all()
    assert (lgdo.explode_cl(lgdo.build_cl(array_exp)) == exp).all()
    # explode
    assert (lgdo.explode(cl, array, array_exp) == array_exp).all()
    assert (lgdo.explode(cl, array) == array_exp).all()
    assert (lgdo.explode([3, 4], [5, 7]) == array_exp).all()
    assert (lgdo.explode(cl, range(len(cl))) == exp).all()
    # explode_arrays
    arrays_out = lgdo.explode_arrays(cl, [array, range(len(cl))])
    assert len(arrays_out) == 2
    assert (arrays_out[0] == array_exp).all()
    assert (arrays_out[1] == exp).all()
    arrays_out = lgdo.explode_arrays(cl, [array, range(len(cl))], arrays_out=arrays_out)
    assert len(arrays_out) == 2
    assert (arrays_out[0] == array_exp).all()
    assert (arrays_out[1] == exp).all()


def test_copy(lgdo_vov):
    assert lgdo_vov == lgdo.copy(lgdo_vov)



from pygama.raw.fc.fc_config_decoder import FCConfigDecoder
from pygama.raw.fc.fc_event_decoder import FCEventDecoder
from pygama.raw.fc.fc_status_decoder import FCStatusDecoder
from pygama.raw.fc.fc_streamer import FCStreamer
from pygama.raw.raw_buffer import RawBuffer, RawBufferList


def test_get_decoder_list():
    streamer = FCStreamer()
    assert isinstance(streamer.get_decoder_list()[0], FCConfigDecoder)
    assert isinstance(streamer.get_decoder_list()[1], FCStatusDecoder)
    assert isinstance(streamer.get_decoder_list()[2], FCEventDecoder)


def test_default_rb_lib(lgnd_test_data):
    streamer = FCStreamer()
    streamer.open_stream(
        lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"), buffer_size=6
    )
    rb_lib = streamer.build_default_rb_lib()
    assert "FCConfigDecoder" in rb_lib.keys()
    assert "FCStatusDecoder" in rb_lib.keys()
    assert "FCEventDecoder" in rb_lib.keys()
    assert rb_lib["FCConfigDecoder"][0].out_name == "FCConfig"
    assert rb_lib["FCStatusDecoder"][0].out_name == "FCStatus"
    assert rb_lib["FCEventDecoder"][0].out_name == "FCEvent"
    assert rb_lib["FCEventDecoder"][0].key_list == range(0, 6)


def test_open_stream(lgnd_test_data):
    streamer = FCStreamer()
    res = streamer.open_stream(
        lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"), buffer_size=6
    )
    assert isinstance(res[0], RawBuffer)
    assert streamer.fcio is not None  # fcio object is instantiated
    assert streamer.packet_id == 0  # packet id is initialized
    assert streamer.n_bytes_read == 11 * 4  # fc header is read
    assert streamer.event_rbkd is not None  # dict containing event info is initialized


def test_read_packet(lgnd_test_data):
    streamer = FCStreamer()
    streamer.open_stream(
        lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"), buffer_size=6
    )
    init_rbytes = streamer.n_bytes_read
    assert streamer.read_packet() is True  # read was successful
    assert streamer.packet_id == 1  # packet id is incremented
    assert streamer.n_bytes_read == init_rbytes + 144 + 2 * streamer.fcio.numtraces * (
        streamer.fcio.nsamples + 3
    )


def test_read_packet_partial(lgnd_test_data):
    streamer = FCStreamer()
    rb_lib = {"FCEventDecoder": RawBufferList()}
    rb_lib["FCEventDecoder"].append(RawBuffer(key_list=range(2, 3), out_name="events"))

    streamer.open_stream(
        lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"),
        rb_lib=rb_lib,
        buffer_size=6,
    )

    assert list(streamer.rb_lib.keys()) == ["FCEventDecoder"]
    assert streamer.rb_lib["FCEventDecoder"][0].key_list == range(2, 3)
    assert streamer.rb_lib["FCEventDecoder"][0].out_name == "events"

    init_rbytes = streamer.n_bytes_read
    assert streamer.read_packet() is True  # read was successful
    assert streamer.packet_id == 1  # packet id is incremented
    assert streamer.n_bytes_read == init_rbytes + 144 + 2 * streamer.fcio.numtraces * (
        streamer.fcio.nsamples + 3
    )


def test_read_chunk(lgnd_test_data):
    streamer = FCStreamer()
    streamer.open_stream(
        lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"), buffer_size=6
    )
    streamer.read_chunk()



"""
The primary function for data conversion into raw-tier LH5 files is
:func:`.build_raw`. This is a one-to many function: one input DAQ file can
generate one or more output raw files. Control of which data ends up in which
files, and in which HDF5 groups inside of each file, is controlled via
:mod:`.raw_buffer` (see below). If no raw buffers specification is specified,
all decoded data should be written to a single output file, with all fields
from each hardware decoder in their own output table.

Currently we support the following DAQ data formats:

* `FlashCam <https://www.mizzi-computer.de/home>`_
* `CoMPASS <https://www.caen.it/products/compass>`_
* `ORCA <https://github.com/unc-enap/Orca>`_, reading out:

  - FlashCam
  - `Struck SIS3302 <https://www.struck.de/sis3302.htm>`_
  - `Struck SIS3316 <https://www.struck.de/sis3316.html>`_
"""

from pygama.raw.build_raw import build_raw

__all__ = ["build_raw"]



from pathlib import Path

import numpy as np
import pytest

from pygama.dsp.errors import DSPFatal
from pygama.dsp.processors import dplms


def test_dplms(compare_numba_vs_python):
    with open(Path(__file__).parent / "dplms_noise_mat.dat") as f:
        nmat = [[float(num) for num in line.split(" ")] for line in f]

    length = 50
    len_wf = 100
    ref = np.zeros(len_wf)
    ref[int(len_wf / 2 - 1) : int(len_wf / 2)] = 1

    # ensure the DSPFatal is raised for a negative length
    with pytest.raises(DSPFatal):
        dplms(-1, nmat, ref, 1, 1, 1, 1, [])

    # ensure the DSPFatal is raised for length not equal to noise matrix shape
    with pytest.raises(DSPFatal):
        dplms(10, nmat, ref, 1, 1, 1, 1, [])

    # ensure the DSPFatal is raised for negative coefficients
    with pytest.raises(DSPFatal):
        dplms(length, nmat, ref, -1, 1, 1, 1, [])
    with pytest.raises(DSPFatal):
        dplms(length, nmat, ref, 1, -1, 1, 1, [])
    with pytest.raises(DSPFatal):
        dplms(length, nmat, ref, 1, 1, -1, 1, [])
    with pytest.raises(DSPFatal):
        dplms(length, nmat, ref, 1, 1, 1, -1, [])
    with pytest.raises(DSPFatal):
        dplms(length, nmat, ref, 1, 1, 1, 2, [])

    dplms_func = dplms(length, nmat, ref, 1, 1, 1, 1, [])

    # ensure to have a valid output
    w_in = np.ones(len_wf)
    w_out = np.empty(len_wf - length + 1)

    assert np.all(compare_numba_vs_python(dplms_func, w_in, w_out))

    # test if there is a nan in w_in
    w_in = np.ones(len_wf)
    w_in[4] = np.nan
    w_out = np.empty(len_wf - length + 1)

    assert np.all(compare_numba_vs_python(dplms_func, w_in, w_out))



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    ["void(float32[:], float32, float32[:])", "void(float64[:], float64, float64[:])"],
    "(n),()->(n)",
    **nb_kwargs,
)
def pole_zero(w_in: np.ndarray, t_tau: float, w_out: np.ndarray) -> None:
    """Apply a pole-zero cancellation using the provided time
    constant to the waveform.

    Parameters
    ----------
    w_in
        the input waveform.
    t_tau
        the time constant of the exponential to be deconvolved.
    w_out
        the pole-zero cancelled waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_pz": {
            "function": "pole_zero",
            "module": "pygama.dsp.processors",
            "args": ["wf_bl", "400*us", "wf_pz"],
            "unit": "ADC"
        }
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(t_tau):
        return

    const = np.exp(-1 / t_tau)
    w_out[0] = w_in[0]
    for i in range(1, len(w_in), 1):
        w_out[i] = w_out[i - 1] + w_in[i] - w_in[i - 1] * const


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32[:])",
        "void(float64[:], float64, float64, float64, float64[:])",
    ],
    "(n),(),(),()->(n)",
    **nb_kwargs,
)
def double_pole_zero(
    w_in: np.ndarray, t_tau1: float, t_tau2: float, frac: float, w_out: np.ndarray
) -> np.ndarray:
    r"""
    Apply a double pole-zero cancellation using the provided time
    constants to the waveform.

    Parameters
    ----------
    w_in
        the input waveform.
    t_tau1
        the time constant of the first exponential to be deconvolved.
    t_tau2
        the time constant of the second exponential to be deconvolved.
    frac
        the fraction which the second exponential contributes.
    w_out
        the pole-zero cancelled waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_pz": {
            "function": "double_pole_zero",
            "module": "pygama.dsp.processors",
            "args": ["wf_bl", "400*us", "20*us", "0.02", "wf_pz"],
            "unit": "ADC"
        }

    Notes
    -----
    This algorithm is an IIR filter to deconvolve the function

    .. math::
        s(t) = A \left[ f \cdot \exp\left(-\frac{t}{\tau_2} \right)
               + (1-f) \cdot \exp\left(-\frac{t}{\tau_1}\right) \right]

    (:math:`f` = `frac`) into a single step function of amplitude :math:`A`.
    This filter is derived by :math:`z`-transforming the input (:math:`s(t)`)
    and output (step function) signals, and then determining the transfer
    function. For shorthand, define :math:`a=\exp(-1/\tau_1)` and
    :math:`b=\exp(-1/\tau_2)`, the transfer function is then:

    .. math::
        H(z) = \frac{1 - (a+b)z^{-1} + abz^{-2}}
                    {1 + (fb - fa - b - 1)z^{-1}-(fb - fa - b)z^{-2}}

    By equating the transfer function to the ratio of output to input waveforms
    :math:`H(z) = w_\text{out}(z) / w_\text{in}(z)` and then taking the
    inverse :math:`z`-transform yields the filter as run in the function, where
    :math:`f` is the `frac` parameter:

    .. math::
        w_\text{out}[n] =& w_\text{in}[n] - (a+b)w_\text{in}[n-1]
                           + abw_\text{in}[n-2] \\
                         & -(fb - fa - b - 1)w_\text{out}[n-1]
                           + (fb - fa - b)w_\text{out}[n-2]
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(t_tau1) or np.isnan(t_tau2) or np.isnan(frac):
        return
    if len(w_in) <= 3:
        raise DSPFatal(
            "The length of the waveform must be larger than 3 for the filter to work safely"
        )

    a = np.exp(-1 / t_tau1)
    b = np.exp(-1 / t_tau2)

    transfer_denom_1 = frac * b - frac * a - b - 1
    transfer_denom_2 = -1 * (frac * b - frac * a - b)
    transfer_num_1 = -1 * (a + b)
    transfer_num_2 = a * b

    w_out[0] = w_in[0]
    w_out[1] = w_in[1]
    w_out[2] = w_in[2]

    for i in range(2, len(w_in), 1):
        w_out[i] = (
            w_in[i]
            + transfer_num_1 * w_in[i - 1]
            + transfer_num_2 * w_in[i - 2]
            - transfer_denom_1 * w_out[i - 1]
            - transfer_denom_2 * w_out[i - 2]
        )



"""
mainly pulser tagging
- gaussian_cut (fits data to a gaussian, returns mean +/- cut_sigma values)
- xtalball_cut (fits data to a crystalball, returns mean +/- cut_sigma values)
- find_pulser_properties (find pulser by looking for which peak has a constant time between events)
- tag_pulsers
"""
import matplotlib.gridspec as gs
import matplotlib.pyplot as plt
import numpy as np
from scipy import stats

from pygama.math.peak_fitting import *


def gaussian_cut(data, cut_sigma=3, plotAxis=None):
    """
    fits data to a gaussian, returns mean +/- cut_sigma values for a cut
    """

    nbins = 100

    median = np.median(data)
    width = np.percentile(data, 80) - np.percentile(data, 20)

    good_data = data[(data > (median - 4 * width)) & (data < (median + 4 * width))]

    hist, bins = np.histogram(good_data, bins=101)  # np.linspace(1,5,101)
    bin_centers = bins[:-1] + (bins[1] - bins[0]) / 2

    # fit gaussians to that
    # result = fit_unbinned(gauss, hist, [median, width/2] )
    # print("unbinned: {}".format(result))

    result = fit_binned(
        gauss,
        hist,
        bin_centers,
        [median, width / 2, np.amax(hist) * (width / 2) * np.sqrt(2 * np.pi)],
    )
    # print("binned: {}".format(result))
    cut_lo = result[0] - cut_sigma * result[1]
    cut_hi = result[0] + cut_sigma * result[1]

    if plotAxis is not None:
        plotAxis.plot(bin_centers, hist, ls="steps-mid", color="k", label="data")
        fit = gauss(bin_centers, *result)
        plotAxis.plot(bin_centers, fit, label="gaussian fit")
        plotAxis.axvline(result[0], color="g", label="fit mean")
        plotAxis.axvline(cut_lo, color="r", label=f"+/- {cut_sigma} sigma")
        plotAxis.axvline(cut_hi, color="r")
        plotAxis.legend()
        # plt.xlabel(params[i])

    return cut_lo, cut_hi, result[0], cut_sigma


def xtalball_cut(data, cut_sigma=3, plotFigure=None):
    """
    fits data to a crystalball, returns mean +/- cut_sigma values for a cut
    """

    nbins = 100

    median = np.median(data)
    width = np.percentile(data, 80) - np.percentile(data, 20)

    good_data = data[(data > (median - 4 * width)) & (data < (median + 4 * width))]

    hist, bins = np.histogram(good_data, bins=101)  # np.linspace(1,5,101)
    bin_centers = bins[:-1] + (bins[1] - bins[0]) / 2

    # fit gaussians to that
    # result = fit_unbinned(gauss, hist, [median, width/2] )
    # print("unbinned: {}".format(result))
    p0 = get_gaussian_guess(hist, bin_centers)
    bounds = [
        (p0[0] * 0.5, p0[1] * 0.5, p0[2] * 0.2, 0, 1),
        (p0[0] * 1.5, p0[1] * 1.5, p0[2] * 5, np.inf, np.inf),
    ]
    result = fit_binned(
        xtalball, hist, bin_centers, [p0[0], p0[1], p0[2], 10, 1], bounds=bounds
    )
    # print("binned: {}".format(result))
    cut_lo = result[0] - cut_sigma * result[1]
    cut_hi = result[0] + cut_sigma * result[1]

    if plotFigure is not None:
        plt.figure(plotFigure.number)
        plt.plot(bin_centers, hist, ls="steps-mid", color="k", label="data")
        fit = xtalball(bin_centers, *result)
        plt.plot(bin_centers, fit, label="xtalball fit")
        plt.axvline(result[0], color="g", label="fit mean")
        plt.axvline(cut_lo, color="r", label=f"+/- {cut_sigma} sigma")
        plt.axvline(cut_hi, color="r")
        plt.legend()
        # plt.xlabel(params[i])

    return cut_lo, cut_hi


def find_pulser_properties(df, energy="trap_max"):
    from .calibration import get_most_prominent_peaks

    # print (df[energy])
    # exit()
    # find pulser by looking for which peak has a constant time between events
    # df should already be grouped by channel

    peak_energies, peak_e_err = get_most_prominent_peaks(df[energy], max_num_peaks=10)
    peak_e_err *= 3

    for e in peak_energies:
        e_cut = (df[energy] > e - peak_e_err) & (df[energy] < e + peak_e_err)
        df_peak = df[e_cut]
        # df_after_0 = df_peak.iloc[1:]
        time_since_last = df_peak.timestamp.values[1:] - df_peak.timestamp.values[:-1]

        tsl = time_since_last[
            (time_since_last >= 0)
            & (time_since_last < np.percentile(time_since_last, 99.9))
        ]
        last_ten = np.percentile(tsl, 97) - np.percentile(tsl, 90)
        first_ten = np.percentile(tsl, 10) - np.percentile(tsl, 3)
        # print("{:e}, {:e}".format(last_ten,first_ten))

        if last_ten > first_ten:
            # print("...no pulser?")
            continue
        else:
            # df["pulser_energy"] = e
            pulser_e = e
            period = stats.mode(tsl).mode[0]

            return pulser_e, peak_e_err, period, energy
    return None


def tag_pulsers(df, chan_info, window=250):
    chan = df.channel.unique()[0]
    df["isPulser"] = 0

    try:
        pi = chan_info.loc[chan]
    except KeyError:
        return df

    energy_name = pi.energy_name
    pulser_energy = pi.pulser_energy
    period = pi.pulser_period
    peak_e_err = pi.peak_e_err

    # pulser_energy, peak_e_err, period, energy_name = chan_info

    e_cut = (df[energy_name] < pulser_energy + peak_e_err) & (
        df[energy_name] > pulser_energy - peak_e_err
    )
    df_pulser = df[e_cut]

    time_since_last = np.zeros(len(df_pulser))
    time_since_last[1:] = (
        df_pulser.timestamp.values[1:] - df_pulser.timestamp.values[:-1]
    )

    # plt.figure()
    # plt.hist(time_since_last, bins=1000)
    # plt.show()

    mode_idxs = (time_since_last > period - window) & (
        time_since_last < period + window
    )

    pulser_events = np.count_nonzero(mode_idxs)
    # print("pulser events: {}".format(pulser_events))
    if pulser_events < 3:
        return df

    df_pulser = df_pulser[mode_idxs]

    ts = df_pulser.timestamp.values
    diff_zero = np.zeros(len(ts))
    diff_zero[1:] = np.around((ts[1:] - ts[:-1]) / period)
    diff_cum = np.cumsum(diff_zero)
    z = np.polyfit(diff_cum, ts, 1)
    p = np.poly1d(z)

    # plt.figure()
    # xp = np.linspace(0, diff_cum[-1])
    # plt.plot(xp,p(xp))
    # plt.scatter(diff_cum,ts)
    # plt.show()

    period = z[0]
    phase = z[1]

    mod = np.abs(df.timestamp - phase) % period

    # pulser_mod  =np.abs(df_pulser.timestamp - phase) %period
    # pulser_mod[ pulser_mod > 10*window] = period - pulser_mod[ pulser_mod > 10*window]
    # plt.hist(pulser_mod , bins="auto")
    # plt.show()
    period_cut = (mod < window) | ((period - mod) < window)

    # print("pulser events: {}".format(np.count_nonzero(e_cut & period_cut)))
    df.loc[e_cut & period_cut, "isPulser"] = 1

    return df



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], int32, int32, float32[:])",
        "void(float64[:], int32, int32, float64[:])",
    ],
    "(n),(),()->(n)",
    **nb_kwargs,
)
def trap_filter(w_in: np.ndarray, rise: int, flat: int, w_out: np.ndarray) -> None:
    """Apply a symmetric trapezoidal filter to the waveform.

    Parameters
    ----------
    w_in
        the input waveform.
    rise
        the number of samples averaged in the rise and fall sections.
    flat
        the delay between the rise and fall sections.
    w_out
        the filtered waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_tf": {
            "function": "trap_filter",
            "module": "pygama.dsp.processors",
            "args": ["wf_pz", "10*us", "3*us", "wf_tf"],
            "unit": "ADC"
        }
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(rise) or np.isnan(flat):
        return

    if int(rise) < 0:
        raise DSPFatal("The number of samples in the rise section must be positive")

    if int(flat) < 0:
        raise DSPFatal("The number of samples in the flat section must be positive")

    if 2 * int(rise) + int(flat) > len(w_in):
        raise DSPFatal("The trapezoid width is wider than the waveform")

    w_out[0] = w_in[0]
    for i in range(1, rise, 1):
        w_out[i] = w_out[i - 1] + w_in[i]
    for i in range(rise, rise + flat, 1):
        w_out[i] = w_out[i - 1] + w_in[i] - w_in[i - rise]
    for i in range(rise + flat, 2 * rise + flat, 1):
        w_out[i] = w_out[i - 1] + w_in[i] - w_in[i - rise] - w_in[i - rise - flat]
    for i in range(2 * rise + flat, len(w_in), 1):
        w_out[i] = (
            w_out[i - 1]
            + w_in[i]
            - w_in[i - rise]
            - w_in[i - rise - flat]
            + w_in[i - 2 * rise - flat]
        )


@guvectorize(
    [
        "void(float32[:], int32, int32, float32[:])",
        "void(float64[:], int32, int32, float64[:])",
    ],
    "(n),(),()->(n)",
    **nb_kwargs,
)
def trap_norm(w_in: np.ndarray, rise: int, flat: int, w_out: np.ndarray) -> None:
    """Apply a symmetric trapezoidal filter to the waveform, normalized by the
    number of samples averaged in the rise and fall sections.

    Parameters
    ----------
    w_in
        the input waveform.
    rise
        the number of samples averaged in the rise and fall sections.
    flat
        the delay between the rise and fall sections.
    w_out
        the normalized, filtered waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_tf": {
            "function": "trap_norm",
            "module": "pygama.dsp.processors",
            "args": ["wf_pz", "10*us", "3*us", "wf_tf"],
            "unit": "ADC"
        }
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(rise) or np.isnan(flat):
        return

    if int(rise) < 0:
        raise DSPFatal("The number of samples in the rise section must be positive")

    if int(flat) < 0:
        raise DSPFatal("The number of samples in the flat section must be positive")

    if 2 * int(rise) + int(flat) > len(w_in):
        raise DSPFatal("The trapezoid width is wider than the waveform")

    w_out[0] = w_in[0] / rise
    for i in range(1, rise, 1):
        w_out[i] = w_out[i - 1] + w_in[i] / rise
    for i in range(rise, rise + flat, 1):
        w_out[i] = w_out[i - 1] + (w_in[i] - w_in[i - rise]) / rise
    for i in range(rise + flat, 2 * rise + flat, 1):
        w_out[i] = (
            w_out[i - 1] + (w_in[i] - w_in[i - rise] - w_in[i - rise - flat]) / rise
        )
    for i in range(2 * rise + flat, len(w_in), 1):
        w_out[i] = (
            w_out[i - 1]
            + (
                w_in[i]
                - w_in[i - rise]
                - w_in[i - rise - flat]
                + w_in[i - 2 * rise - flat]
            )
            / rise
        )


@guvectorize(
    [
        "void(float32[:], int32, int32, int32, float32[:])",
        "void(float64[:], int32, int32, int32, float64[:])",
    ],
    "(n),(),(),()->(n)",
    **nb_kwargs,
)
def asym_trap_filter(
    w_in: np.ndarray, rise: int, flat: int, fall: int, w_out: np.ndarray
) -> None:
    """Apply an asymmetric trapezoidal filter to the waveform, normalized by
    the number of samples averaged in the rise and fall sections.

    Parameters
    ----------
    w_in
        the input waveform.
    rise
        the number of samples averaged in the rise section.
    flat
        the delay between the rise and fall sections.
    fall
        the number of samples averaged in the fall section.
    w_out
        the normalized, filtered waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_af": {
            "function": "asym_trap_filter",
            "module": "pygama.dsp.processors",
            "args": ["wf_pz", "128*ns", "64*ns", "2*us", "wf_af"],
            "unit": "ADC"
        }
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(rise) or np.isnan(flat) or np.isnan(fall):
        return

    if int(rise) < 0:
        raise DSPFatal("The number of samples in the rise section must be positive")

    if int(flat) < 0:
        raise DSPFatal("The number of samples in the flat section must be positive")

    if int(fall) < 0:
        raise DSPFatal("The number of samples in the fall section must be positive")

    if int(rise) + int(flat) + int(fall) > len(w_in):
        raise DSPFatal("The trapezoid width is wider than the waveform")

    w_out[0] = w_in[0] / rise
    for i in range(1, rise, 1):
        w_out[i] = w_out[i - 1] + w_in[i] / rise
    for i in range(rise, rise + flat, 1):
        w_out[i] = w_out[i - 1] + (w_in[i] - w_in[i - rise]) / rise
    for i in range(rise + flat, rise + flat + fall, 1):
        w_out[i] = (
            w_out[i - 1]
            + (w_in[i] - w_in[i - rise]) / rise
            - w_in[i - rise - flat] / fall
        )
    for i in range(rise + flat + fall, len(w_in), 1):
        w_out[i] = (
            w_out[i - 1]
            + (w_in[i] - w_in[i - rise]) / rise
            - (w_in[i - rise - flat] - w_in[i - rise - flat - fall]) / fall
        )


@guvectorize(
    [
        "void(float32[:], int32, int32, float32, float32[:])",
        "void(float64[:], int32, int32, float64, float64[:])",
    ],
    "(n),(),(),()->()",
    **nb_kwargs,
)
def trap_pickoff(
    w_in: np.ndarray, rise: int, flat: int, t_pickoff: float, a_out: float
) -> None:
    """Pick off the value at the provided index of a symmetric trapezoidal
    filter to the input waveform, normalized by the number of samples averaged
    in the rise and fall sections.

    Parameters
    ----------
    w_in
        the input waveform.
    rise
        the number of samples averaged in the rise and fall sections.
    flat
        the delay between the rise and fall sections.
    t_pickoff
        the waveform index to pick off.
    a_out
        the output pick-off value of the filtered waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "ct_corr": {
            "function": "trap_pickoff",
            "module": "pygama.dsp.processors",
            "args": ["wf_pz", "1.5*us", 0, "tp_0", "ct_corr"],
            "unit": "ADC"
        }
    """
    a_out[0] = np.nan

    if np.isnan(w_in).any() or np.isnan(rise) or np.isnan(flat) or np.isnan(t_pickoff):
        return

    if np.floor(t_pickoff) != t_pickoff:
        raise DSPFatal("The pick-off index must be an integer")

    if int(rise) < 0:
        raise DSPFatal("The number of samples in the rise section must be positive")

    if int(flat) < 0:
        raise DSPFatal("The number of samples in the flat section must be positive")

    if 2 * int(rise) + int(flat) > len(w_in):
        raise DSPFatal("The trapezoid width is wider than the waveform")

    i_1 = 0.0
    i_2 = 0.0
    start_time = int(t_pickoff + 1)

    if not len(w_in) >= start_time >= 2 * rise + flat:
        return

    for i in range(start_time - rise, start_time, 1):
        i_1 += w_in[i]
    for i in range(start_time - 2 * rise - flat, start_time - rise - flat, 1):
        i_2 += w_in[i]
    a_out[0] = (i_1 - i_2) / rise



from __future__ import annotations

import logging

import fcutils

from pygama.raw.data_decoder import DataDecoder
from pygama.raw.data_streamer import DataStreamer
from pygama.raw.fc.fc_config_decoder import FCConfigDecoder
from pygama.raw.fc.fc_event_decoder import FCEventDecoder
from pygama.raw.fc.fc_status_decoder import FCStatusDecoder
from pygama.raw.raw_buffer import RawBuffer, RawBufferLibrary

log = logging.getLogger(__name__)


class FCStreamer(DataStreamer):
    """
    Decode FlashCam data, using the ``fcutils`` package to handle file access,
    and the FlashCam data decoders to save the results and write to output.
    """

    def __init__(self) -> None:
        super().__init__()
        self.fcio = None
        self.config_decoder = FCConfigDecoder()
        self.status_decoder = FCStatusDecoder()
        self.event_decoder = FCEventDecoder()
        self.event_tables = {}
        self.event_rbkd = None
        self.status_rb = None

    def get_decoder_list(self) -> list[DataDecoder]:
        dec_list = []
        dec_list.append(self.config_decoder)
        dec_list.append(self.status_decoder)
        dec_list.append(self.event_decoder)
        return dec_list

    def open_stream(
        self,
        fcio_filename: str,
        rb_lib: RawBufferLibrary = None,
        buffer_size: int = 8192,
        chunk_mode: str = "any_full",
        out_stream: str = "",
    ) -> list[RawBuffer]:
        """Initialize the FlashCam data stream.

        Refer to the documentation for
        :meth:`.raw.data_streamer.DataStreamer.open_stream` for a description
        of the parameters.

        Returns
        -------
        header_data
            a list of length 1 containing the raw buffer holding the
            :class:`~.fc_config_decoder.FCConfig` table.
        """
        self.fcio = fcutils.fcio(fcio_filename)
        self.n_bytes_read = 0

        # read in file header (config) info
        fc_config = self.config_decoder.decode_config(
            self.fcio
        )  # returns an lgdo.Struct
        self.event_decoder.set_file_config(fc_config)
        self.n_bytes_read += 11 * 4  # there are 11 ints in the fcio_config struct

        # initialize the buffers in rb_lib. Store them for fast lookup
        super().open_stream(
            fcio_filename,
            rb_lib,
            buffer_size=buffer_size,
            chunk_mode=chunk_mode,
            out_stream=out_stream,
        )
        if rb_lib is None:
            rb_lib = self.rb_lib

        # get the status rb_list and pull out its first element
        status_rb_list = (
            rb_lib["FCStatusDecoder"] if "FCStatusDecoder" in rb_lib else None
        )
        if status_rb_list is not None:
            if len(status_rb_list) != 1:
                log.warning(
                    f"status_rb_list had length {len(status_rb_list)}, ignoring all but the first"
                )
            if len(status_rb_list) == 0:
                self.status_rb = None
            else:
                self.status_rb = status_rb_list[0]
        self.event_rbkd = (
            rb_lib["FCEventDecoder"].get_keyed_dict()
            if "FCEventDecoder" in rb_lib
            else None
        )

        # set up data loop variables
        self.packet_id = 0  # for storing packet order in output tables

        if "FCConfigDecoder" in rb_lib:
            config_rb_list = rb_lib["FCConfigDecoder"]
            if len(config_rb_list) != 1:
                log.warning(
                    f"config_rb_list had length {len(config_rb_list)}, ignoring all but the first"
                )
            rb = config_rb_list[0]
        else:
            rb = RawBuffer(lgdo=fc_config)
        rb.loc = 1  # we have filled this buffer
        return [rb]

    def close_stream(self) -> None:
        self.fcio = None  # should cause close file in fcio.__dealloc__

    def read_packet(self) -> bool:
        rc = self.fcio.get_record()
        if rc == 0:
            return False  # no more data

        self.packet_id += 1

        if rc == 1:  # config (header) data
            log.warning(
                f"got a header after start of run? n_bytes_read = {self.n_bytes_read}"
            )
            self.n_bytes_read += 11 * 4  # there are 11 ints in the fcio_config struct
            return True

        elif rc == 2:  # calib record -- no longer supported
            log.warning(
                f"warning: got a calib record? n_bytes_read = {self.n_bytes_read}"
            )
            return True

        # FIXME: push to a buffer of skipped packets?
        # FIXME: need to at least update n_bytes?
        elif rc == 5:  # recevent
            log.warning(
                f"got a RecEvent packet -- skipping? n_bytes_read = {self.n_bytes_read}"
            )
            # sizeof(fcio_recevent): (6 + 3*10 + 1*2304 + 3*4000)*4
            self.n_bytes_read += 57360
            return True

        # Status record
        elif rc == 4:
            if self.status_rb is not None:
                self.any_full |= self.status_decoder.decode_packet(
                    self.fcio, self.status_rb, self.packet_id
                )
            # sizeof(fcio_status): (3 + 10 + 256*(10 + 9 + 16 + 4 + 256))*4
            self.n_bytes_read += 302132
            return True

        # Event or SparseEvent record
        elif rc == 3 or rc == 6:
            if self.event_rbkd is not None:
                self.any_full |= self.event_decoder.decode_packet(
                    self.fcio, self.event_rbkd, self.packet_id
                )
            # sizeof(fcio_event): (5 + 3*10 + 1)*4 + numtraces*(1 + nsamples+2)*2
            self.n_bytes_read += (
                144 + self.fcio.numtraces * (self.fcio.nsamples + 3) * 2
            )
            return True



import copy
import inspect
import os
import re
import shutil
import uuid
from getpass import getuser
from pathlib import Path
from tempfile import gettempdir

import numpy as np
import pytest
from legendtestdata import LegendTestData

from pygama.dsp import build_dsp
from pygama.raw import build_raw

config_dir = Path(__file__).parent / "dsp" / "configs"
_tmptestdir = os.path.join(
    gettempdir(), "pygama-tests-" + getuser() + str(uuid.uuid4())
)


@pytest.fixture(scope="session")
def tmptestdir():
    os.mkdir(_tmptestdir)
    yield _tmptestdir
    shutil.rmtree(_tmptestdir)


def pytest_sessionfinish(session, exitstatus):
    if exitstatus is True:
        os.rmdir(_tmptestdir)


@pytest.fixture(scope="session")
def lgnd_test_data():
    ldata = LegendTestData()
    ldata.checkout("c089a59")
    return ldata


@pytest.fixture(scope="session")
def dsp_test_file(lgnd_test_data):
    out_name = "/tmp/LDQTA_r117_20200110T105115Z_cal_geds_dsp.lh5"
    build_dsp(
        lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"),
        out_name,
        dsp_config=f"{config_dir}/icpc-dsp-config.json",
        database={"pz": {"tau": 27460.5}},
        write_mode="r",
    )
    assert os.path.exists(out_name)

    return out_name


@pytest.fixture(scope="session")
def multich_raw_file(lgnd_test_data):
    out_file = "/tmp/L200-comm-20211130-phy-spms.lh5"
    out_spec = {
        "FCEventDecoder": {
            "ch{key}": {
                "key_list": [[0, 6]],
                "out_stream": out_file + ":{name}",
                "out_name": "raw",
            }
        }
    }

    build_raw(
        in_stream=lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"),
        out_spec=out_spec,
        overwrite=True,
    )
    assert os.path.exists(out_file)

    return out_file


@pytest.fixture(scope="session")
def dsp_test_file_spm(multich_raw_file):
    chan_config = {
        "ch0/raw": f"{config_dir}/sipm-dsp-config.json",
        "ch1/raw": f"{config_dir}/sipm-dsp-config.json",
        "ch2/raw": f"{config_dir}/sipm-dsp-config.json",
    }

    out_file = "/tmp/L200-comm-20211130-phy-spms_dsp.lh5"
    build_dsp(
        multich_raw_file,
        out_file,
        {},
        n_max=5,
        lh5_tables=chan_config.keys(),
        chan_config=chan_config,
        write_mode="r",
    )

    assert os.path.exists(out_file)

    return out_file


@pytest.fixture(scope="session")
def compare_numba_vs_python():
    def numba_vs_python(func, *inputs):
        """
        Function for testing that the numba and python versions of a
        function are equal.

        Parameters
        ----------
        func
            The Numba-wrapped function to be tested
        *inputs
            The various inputs to be passed to the function to be
            tested.

        Returns
        -------
        func_output
            The output of the function to be used in a unit test.

        """

        if "->" in func.signature:
            # parse outputs from function signature
            all_params = list(inspect.signature(func).parameters)
            output_sizes = re.findall(r"(\(n*\))", func.signature.split("->")[-1])
            noutputs = len(output_sizes)
            output_names = all_params[-noutputs:]

            # numba outputs
            outputs_numba = func(*inputs)
            if noutputs == 1:
                outputs_numba = [outputs_numba]

            # unwrapped python outputs
            func_unwrapped = inspect.unwrap(func)
            output_dict = {key: np.empty(len(inputs[0])) for key in output_names}
            func_unwrapped(*inputs, **output_dict)
            for spec, key in zip(output_sizes, output_dict):
                if spec == "()":
                    output_dict[key] = output_dict[key][0]
            outputs_python = [output_dict[key] for key in output_names]
        else:
            # we are testing a factory function output, which updates
            # a single output in-place
            noutputs = 1
            # numba outputs
            func(*inputs)
            outputs_numba = copy.deepcopy(inputs[-noutputs:])

            # unwrapped python outputs
            func_unwrapped = inspect.unwrap(func)
            func_unwrapped(*inputs)
            outputs_python = copy.deepcopy(inputs[-noutputs:])

        # assert that numba and python are the same up to floating point
        # precision, setting nans to be equal
        assert np.allclose(outputs_numba, outputs_python, equal_nan=True)

        # return value for comparison with expected solution
        return outputs_numba[0] if noutputs == 1 else outputs_numba

    return numba_vs_python



import os
from pathlib import Path

import numpy as np

import pygama.lgdo.lh5_store as store
from pygama.dsp import build_dsp

config_dir = Path(__file__).parent / "configs"
dsp_file = "/tmp/LDQTA_r117_20200110T105115Z_cal_geds__numpy_test_dsp.lh5"


def test_build_dsp(lgnd_test_data):
    build_dsp(
        f_raw=lgnd_test_data.get_path(
            "lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"
        ),
        f_dsp=dsp_file,
        dsp_config=f"{config_dir}/numpy-parsing.json",
        write_mode="r",
    )
    assert os.path.exists(dsp_file)


def test_numpy_math_constants_dsp():
    df = store.load_nda(dsp_file, ["timestamp", "calc1", "calc2", "calc3"], "geds/dsp/")

    a1 = df["timestamp"] - df["timestamp"] - np.pi * df["timestamp"]
    a2 = df["timestamp"] - df["timestamp"] - np.pi
    a3 = df["timestamp"] - df["timestamp"] - np.pi * np.e

    f1 = df["calc1"]
    f2 = df["calc2"]
    f3 = df["calc3"]

    assert (a1 == f1).all()
    assert (a2 == f2).all()
    assert (a3 == f3).all()


def test_numpy_infinity_and_nan_dsp():
    df = store.load_nda(dsp_file, ["calc4", "calc5", "calc6"], "geds/dsp/")

    assert (np.isnan(df["calc4"])).all()
    assert (np.isneginf(df["calc5"])).all()
    assert (np.isnan(df["calc6"])).all()



import numpy as np

from pygama.pargen import energy_cal


def test_peak_match():
    expected_peaks_kev = [1460, 2614.5]
    peaks_adu = [
        78.676315,
        288.4798,
        603.18506,
        1337.4973,
        2019.3586,
        3225.7288,
        3907.59,
        5795.8213,
        8470.816,
    ]
    pars, best_ixtup, best_iytup = energy_cal.poly_match(
        peaks_adu, expected_peaks_kev, deg=0, atol=10
    )
    assert np.array_equal(best_ixtup, [5, 7])



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    ["void(float32[:], float32, float32[:])", "void(float64[:], float64, float64[:])"],
    "(n),(),(m)",
    **nb_kwargs,
)
def windower(w_in: np.ndarray, t0_in: int, w_out: np.ndarray) -> None:
    """Return a shorter sample of the waveform, starting at the
    specified index.

    Note
    ----
    The length of the output waveform is determined by the length of `w_out`
    rather than an input parameter.  If the the length of `w_out` plus `t0_in`
    extends past the end of `w_in` or if `t0_in` is negative, remaining values
    are padded with :any:`numpy.nan`.

    Parameters
    ----------
    w_in
        the input waveform.
    t0_in
        the starting index of the window.
    w_out
        the windowed waveform.
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(t0_in):
        return

    if len(w_out) >= len(w_in):
        raise DSPFatal("The windowed waveform must be smaller than the input waveform")

    beg = min(int(t0_in), len(w_in))
    end = max(beg + len(w_out), 0)
    if beg < 0:
        w_out[: len(w_out) - end] = np.nan
        w_out[len(w_out) - end :] = w_in[:end]
    elif end < len(w_in):
        w_out[:] = w_in[beg:end]
    else:
        w_out[: len(w_in) - beg] = w_in[beg : len(w_in)]
        w_out[len(w_in) - beg :] = np.nan
